#include <sys/types.h>
#include <openssl/blowfish.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include <err.h>
#include <stdio.h>
#include <string.h>
#include <sysexits.h>

#include "vncryptctl.h"

#define DEFAULT_KEY_LENGTH 32

struct bf_cbc_keydata {
	u_int32_t iv[2];		/* IV */
	size_t len;			/* key length, variable */
	char key[0];
};

static check_params_t check_bf_cbc;
static generate_t generate_bf_cbc;
static create_context_t create_ctx;
static clear_context_t clear_ctx;
static decrypt_block_t decrypt_block;

static struct alg_params bf_cbc_params = {
	ALG_BF_CBC,
	"bf-cbc",
	"blowfish in single-key CBC mode",
	check_bf_cbc,
	generate_bf_cbc,
	create_ctx,
	clear_ctx,
	decrypt_block
};

static void
check_bf_cbc(u_long *key_length, u_long *fake) {
	if (*key_length == 0)
		*key_length = DEFAULT_KEY_LENGTH;

	if (*key_length > 56)
		errx(EX_USAGE,
		     "invalid blowfish key length, must be <= 56");
	if (*fake != 0)
		errx(EX_USAGE, "bf-cbc has no subkeys");
}

static void
generate_bf_cbc(char **pptr, size_t *psize, u_long key_length, 
		u_long fake __unused) {
	struct bf_cbc_keydata *pkey;

	LOG(("generating %ld-bit bf_cbc key\n", key_length * 8));
	
	*psize = sizeof(struct bf_cbc_keydata) + key_length;

	*pptr = malloc(*psize);
	if (*pptr == NULL)
		err(EX_TEMPFAIL, "can't allocate %d bytes", *psize);

	pkey = (struct bf_cbc_keydata *)*pptr;
	pkey->len = key_length;

	if (RAND_bytes((u_char *)pkey->iv, sizeof(pkey->iv)) != 1)
		errx(EX_TEMPFAIL, "can't get %d random bytes: %s",
		     sizeof(pkey->iv),
		     ERR_error_string(ERR_get_error(), NULL));

	if (RAND_bytes((u_char *)(pkey + 1), key_length) != 1)
		errx(EX_TEMPFAIL, "can't get %ld random bytes: %s",
		     key_length,
		     ERR_error_string(ERR_get_error(), NULL));
}

void
register_bf_cbc(void) {
	register_alg(&bf_cbc_params);
}

/* don't care about multiple contexts -- we don't use them */

struct bf_cbc_key {
	u_int32_t seed[2];
	BF_KEY key;
} cryptocontext;

static void *
create_ctx(const void *ptr, size_t size) {
	const struct bf_cbc_keydata *data = (const struct bf_cbc_keydata *)ptr;

	if (size != sizeof(*data) + data->len)
		errx(EX_DATAERR, "keyfile broken");

	cryptocontext.seed[0] = data->iv[0];
	cryptocontext.seed[1] = data->iv[1];
	BF_set_key(&cryptocontext.key, data->len, data->key);

	if (verbose > 1) {
		u_int i;

		printf("seed: %08x %08x\n", data->iv[0], data->iv[1]);
		printf("key: ");
		for (i = 0; i < data->len; i++)
			printf("%02x ", data->key[i] & 0xff);
		printf("\n");
	}

	return &cryptocontext;
}

static void
clear_ctx(void *ctx) {
	bzero(ctx, sizeof(struct bf_cbc_key));
}

static void
checksum(u_int32_t sum[2], const u_int32_t *base, const size_t len) {
	/* simple checksum, simular to ADLER32 */
	u_int i;

        for (i = 0; i < len / sizeof(*base); i++) {
		sum[0] += base[i];
		sum[1] += sum[0];
        }
}

static void
decrypt_block(void *ctx, u_long blockno, void *buf, size_t size) {
	struct bf_cbc_key *key = (struct bf_cbc_key *)ctx;
	u_char *data = (u_char *)buf;
	u_long *tmp = (u_long *)buf;
	u_long iv[2];
	u_char tmpbuf[size];
	u_int i;

	/* bf_cbc expects data in network byte order */
	for (i = 0; i < size / sizeof(*tmp); i++)
		tmp[i] = htonl(tmp[i]);
	
	/* decrypting most of the block */
	iv[0] = tmp[0];
	iv[1] = tmp[1];
	
	BF_cbc_encrypt(data + 8, tmpbuf + 8, size - 8, &key->key, (u_char *)iv, BF_DECRYPT);

	/* converting back to host byte order for checksum calculation */
	tmp = (u_long *)tmpbuf;
	
	for (i = 2; i < size / sizeof(*tmp); i++)
		tmp[i] = ntohl(tmp[i]);

	/* generating checksum */
	iv[0] = key->seed[0] ^ blockno;
	iv[1] = key->seed[1] ^ blockno;

	checksum((u_int32_t *)iv, (u_int32_t *)(tmpbuf + 8), size - 8);

	/* decrypting first 8 bytes */
	iv[0] = htonl(iv[0]);
	iv[1] = htonl(iv[1]);

	BF_cbc_encrypt(data, tmpbuf, 8, &key->key, (u_char *)iv, BF_DECRYPT);

	/* converting them to host byteorder too */
	for (i = 0; i < 2; i++)
		tmp[i] = ntohl(tmp[i]);

	/* copying to original buffer */
	bcopy(tmpbuf, buf, size);
}
