#include <sys/types.h>
#include <openssl/blowfish.h>
#include <openssl/hmac.h>
#include <err.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include "vncryptctl.h"

static __dead2 void
show_help(void) {
	const char **names = get_alg_names();
	const char **comments = get_alg_comments();
	int n = 0;

	printf("Usage: vncryptctl makekey [options] keyfile\n");
	printf("Valid options:\n"
	       "\t-h\tthis help\n"
	       "\t-a alg\tcipher for use\n"
	       "\t-l len\tkey length for algorithms with "
	       "variable key length\n"
	       "\t-s len\tsubkey length for algorithms with "
	       "secondary keys\n");

	printf("Avaliable ciphers:\n");
	while (names[n]) {
		printf("%s\t%s\n", names[n], comments[n]);
		n++;
	}
	
	exit(0);
}

void
generate_key(int argc, char **argv) {
	struct open_header openhdr = { KEYFILE_SIG, 0, {} };
	struct secret_header sechdr = { 0 };
	struct alg_params *alg = NULL;
	HMAC_CTX hmac;
	BF_KEY master_key;
	char keybuf[MASTER_KEY_LEN + BF_BLOCK];
	u_long key_length = 0;
	u_long subkey_length = 0;
	int n = 0;
	int ch;
	int fd;
	char *ptr;

	optreset = 1;
	optind = 0;
	
	/* parsing options */
	
	while ((ch = getopt(argc, argv, "ha:l:s:")) != -1)
		switch (ch) {
		case 'a':
			/* choosing cipher */
			if (alg != NULL)
				errx(EX_USAGE, "algorithm already selected");
			
			alg = get_alg_by_name(optarg);
			if (alg == NULL) {
				warnx("unknown algorithm: %s\n", optarg);
				show_help();
			}
			break;
		{
			char *lastchar;
		case 'l':
			if (key_length != 0)
				errx(EX_USAGE, "key length already selected");
			key_length = strtoul(optarg, &lastchar, 10);
			if (*lastchar != '\0' || key_length == ULONG_MAX ||
			    key_length == 0)
				errx(EX_USAGE, "invalid key length `%s'",
				     optarg);
			break;
		case 's':
			if (subkey_length != 0)
				errx(EX_USAGE, "subkey length already "
						"selected");
			subkey_length = strtoul(optarg, &lastchar, 10);
			if (*lastchar != '\0' || subkey_length == ULONG_MAX ||
			    subkey_length == 0)
				errx(EX_USAGE, "invalid subkey length `%s'",
				     optarg);
			break;
		}
		case 'h': /* FALL-THROUGH */
		default:
			show_help();
		}

	argc -= optind;
	argv += optind;

	if (argc != 1) {
		warnx("keyfile not specified");
		show_help();
	}

	/* setting defaults and checking params */

	if (alg == NULL)
		alg = get_first_alg();

	sechdr.alg_id = alg->id;

	alg->check_func(&key_length, &subkey_length);

	/* generating key */
	
	alg->generate_func(&ptr, &openhdr.keydata_len, key_length,
		       subkey_length);
	
	/* generating master key */
	
	get_key_from_passphrase(keybuf, sizeof(keybuf));

	/* encrypting data */

	BF_set_key(&master_key, MASTER_KEY_LEN, &(keybuf[BF_BLOCK]));

	BF_cfb64_encrypt((u_char *)&sechdr, (u_char *)&sechdr, sizeof(sechdr),
			 &master_key, keybuf, &n, BF_ENCRYPT);

	BF_cfb64_encrypt(ptr, ptr, openhdr.keydata_len, &master_key,
			 keybuf, &n, BF_ENCRYPT);

	/* cleaning master key */

	bzero(&master_key, sizeof(master_key));

	/* generating MAC on secret_header and keydata */

	HMAC_Init(&hmac, &(keybuf[BF_BLOCK]), MASTER_KEY_LEN, EVP_md5());

	bzero(keybuf, sizeof(keybuf));

	HMAC_Update(&hmac, (char *)&sechdr, sizeof(sechdr));
	HMAC_Update(&hmac, ptr, openhdr.keydata_len);
	HMAC_Final(&hmac, openhdr.mac, NULL);
	HMAC_cleanup(&hmac);

	/* writing keyfile */

	fd = open(argv[0], O_WRONLY | O_CREAT | O_TRUNC, 0600);
	if (fd == -1)
		err(EX_CANTCREAT, "can't open output file `%s'", argv[0]);

	if (write(fd, &openhdr, sizeof(openhdr)) != (int)sizeof(openhdr))
		err(EX_CANTCREAT, "can't write to keyfile");

	if (write(fd, &sechdr, sizeof(sechdr)) != (int)sizeof(sechdr))
		err(EX_CANTCREAT, "can't write to keyfile");

	if (write(fd, ptr, openhdr.keydata_len) != (int)openhdr.keydata_len)
		err(EX_CANTCREAT, "can't write to keyfile");

	close(fd);

	/* cleaning */

	bzero(ptr, openhdr.keydata_len);

	free(ptr);
	
	return ;
}
