#include <sys/types.h>
#include <openssl/blowfish.h>
#include <openssl/hmac.h>
#include <openssl/md5.h>
#include <fcntl.h>
#include <err.h>
#include <pwd.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include "vncryptctl.h"

char prng_iv[BF_BLOCK] = { 0xb0, 0x3b, 0xfe, 0x1b, 0xda, 0xce, 0xac, 0x2b };

void get_key_from_passphrase(void *buf, size_t size) {
	char *pass;
	char hash[MD5_DIGEST_LENGTH];
	BF_KEY prng_key;
	int n = 0;

	if (passfd == -1) {
		pass = getpass("Enter keyfile passphrase:");
	} else {
		pass = malloc(_PASSWORD_LEN + 1);
		if (pass == NULL)
			err(EX_OSERR, "malloc() failed");
		bzero(pass, _PASSWORD_LEN + 1);
		if (read(passfd, pass, _PASSWORD_LEN) <= 0)
			err(EX_NOINPUT, "no passphrase");
	}

	/* generating PRNG key by hashing passphrase */

	MD5(pass, strlen(pass), hash);
	bzero(pass, _PASSWORD_LEN);
	if (passfd != -1)
		free(pass);

	/* loading PRNG key to cipher */
	
	BF_set_key(&prng_key, MD5_DIGEST_LENGTH, hash);
	bzero(hash, sizeof(hash));

	/* generating requested key */

	bzero(buf, size);
	BF_cfb64_encrypt(buf, buf, size, &prng_key, prng_iv, &n, BF_ENCRYPT);

	/* clearing key */

	bzero(&prng_key, sizeof(prng_key));
}

void
read_keyfile(const char *filename, struct vncr_generic_params **param_ptr,
	 size_t *keydata_len, u_int *algid) {
	int fd;
	struct open_header openhdr;
	struct secret_header sechdr;
	struct vncr_generic_params *params;
	BF_KEY master_key;
	HMAC_CTX hmac;
	char keybuf[MASTER_KEY_LEN + BF_BLOCK];
	char real_mac[MD5_LBLOCK];
	int n = 0;

	/* loading open header */

	fd = open(filename, O_RDONLY);
	if (fd == -1)
		err(EX_NOINPUT, "can't open keyfile `%s'", filename);

	if (read(fd, &openhdr, sizeof(openhdr)) != (int)sizeof(openhdr))
		err(EX_NOINPUT, "can't read keyfile");

	/* verifying signature */
	
	if (openhdr.sig != KEYFILE_SIG)
		errx(EX_DATAERR, "`%s' is not valid keyfile", filename);

	/* loading secret header and data */

	if (read(fd, &sechdr, sizeof(sechdr)) != (int)sizeof(sechdr))
		err(EX_NOINPUT, "can't read keyfile");

	*param_ptr = params = malloc(sizeof(*params) + openhdr.keydata_len);

	if (params == NULL)
		err(EX_TEMPFAIL, "can't allocate keydata buffer");

	if (read(fd, params + 1, openhdr.keydata_len)
	    != (int)openhdr.keydata_len)
		err(EX_NOINPUT, "can't read keyfile");

	close(fd);

	/* checking mac */

	get_key_from_passphrase(keybuf, sizeof(keybuf));

	HMAC_Init(&hmac, &(keybuf[BF_BLOCK]), MASTER_KEY_LEN, EVP_md5());
	HMAC_Update(&hmac, (char *)&sechdr, sizeof(sechdr));
	HMAC_Update(&hmac, (char *)(params + 1), openhdr.keydata_len);
	HMAC_Final(&hmac, real_mac, NULL);
	HMAC_cleanup(&hmac);

	if (bcmp(openhdr.mac, real_mac, MD5_LBLOCK)) {
		bzero(keybuf, sizeof(keybuf));
		errx(EX_DATAERR, "invalid passphrase");
	}

	BF_set_key(&master_key, MASTER_KEY_LEN, &(keybuf[BF_BLOCK]));

	BF_cfb64_encrypt((u_char *)&sechdr, (u_char *)&sechdr, sizeof(sechdr),
			 &master_key, keybuf, &n, BF_DECRYPT);

	BF_cfb64_encrypt((u_char *)(params + 1), (u_char *)(params + 1),
			 openhdr.keydata_len, &master_key,
			 keybuf, &n, BF_DECRYPT);

	/* using our data */

	*algid = sechdr.alg_id;
	*keydata_len = openhdr.keydata_len;

	/* cleanup */
	
	bzero(&master_key, sizeof(master_key));
	bzero(keybuf, sizeof(keybuf));
	bzero(&sechdr, sizeof(sechdr));
}
