/*
 * Copyright (c) 2002 Andrey Sverdlichenko
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by 
 *	Andrey Sverdlichenko
 * 4. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Id: main.c,v 1.2 2003/01/08 19:17:43 rblaze Exp $ */
/* User-level decryption of disk image.
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <openssl/blowfish.h>
#include <openssl/hmac.h>
#include <err.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include <dev/vncrypt/vn_crypt.h>
#include <machine/param.h>

#include "vncryptctl.h"

int verbose = 0;
int passfd = -1;

static __dead2 void
show_help(void) {
	printf("Usage: vndecrypt [options] source dest keyfile\n");
	printf("Valid options:\n"
	       "\t-p fd\tread passphrase from this descriptor\n"
	       "\t-h\tthis help\n"
	       "\t-v\tverbose mode\n");
	exit(0);
}

int
main(int argc, char **argv) {
	struct vncr_generic_params *params;
	size_t keydata_len;
	u_int algid;
	int ch;
	int ifd, ofd;
	char buf[DEV_BSIZE];
	u_int blocks = 0;
	void *context;
	struct alg_params *alg;

	register_bf_cbc();
	
	if (argc == 1)
		show_help();

	/* parsing options */

	while ((ch = getopt(argc, argv, "vhp:")) != -1)
		switch (ch) {
		case 'v': verbose++;
			  break;
		{
			char *lastchar;
		case 'p':
			if (passfd != -1)
				errx(EX_USAGE,
				     "passphrase fd already selected");
			passfd = strtol(optarg, &lastchar, 10);
			if (*lastchar != '\0' || passfd == LONG_MAX
			    || passfd < 0)
				errx(EX_USAGE, "invalid fd `%s'", optarg);
			break;
		}
		case 'h': /* FALL-THROUGH */
		default:
			show_help();
		}

	argc -= optind;
	argv += optind;

	if (argc != 3) {
		warnx("keyfile or data files not specified");
		show_help();
	}

	read_keyfile(argv[2], &params, &keydata_len, &algid);

	alg = get_alg_by_id(algid);
	if (alg == NULL)
		errx(EX_DATAERR, "unknown algorithm id %d", algid);
	
	if (alg->create_ctx_func == NULL
	    || alg->clear_ctx_func == NULL
	    || alg->decrypt_func == NULL)
		errx(EX_SOFTWARE, "alg %s has not user-level "
				"decryption support", alg->name);

	LOG(("decrypting from %s to %s with alg %s\n", argv[0], argv[1],
				alg->name));

	/* setting key */

	context = alg->create_ctx_func(params + 1, keydata_len);

	/* ok, now decrypting device */

	ifd = open(argv[0], O_RDONLY);
	if (ifd == -1)
		err(EX_NOINPUT, "can't open `%s'", argv[0]);

	ofd = open(argv[1], O_WRONLY | O_CREAT | O_TRUNC, 0400);
	if (ofd == -1)
		err(EX_NOINPUT, "can't open `%s'", argv[1]);

	while (read(ifd, buf, sizeof(buf)) == sizeof(buf)) {
		alg->decrypt_func(context, blocks++, buf, sizeof(buf));

		if (write(ofd, buf, sizeof(buf)) != sizeof(buf))
			err(EX_IOERR, "write failed");
	}

	close(ifd);
	close(ofd);

	/* cleanup */

	alg->clear_ctx_func(context);
	bzero(params, sizeof(*params) + keydata_len);
	free(params);

	LOG(("%d blocks processed\n", blocks));

	return 0;
}
