/*
 * Copyright (c) 2001 Andrey Sverdlichenko
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by 
 *	Andrey Sverdlichenko
 * 4. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Id: vn_aes_cbc.c,v 1.3 2002/02/25 21:51:52 rblaze Exp $ */
/* AES (Rijndael) in CBC mode. IV generated from block number and contents.
 * Expected to be secure.
 * Potential problems:
 * - all disk encrypted on ONE key.
 * - IV uses all but first 8 bytes of block. This will be fixed later.
 */
#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <dev/vncrypt/vn_crypt.h>
#include <crypto/rijndael/rijndael.h>

#if BITSPERBLOCK != 128
#error Rijndael default block length changed
#endif

MALLOC_DEFINE(M_VNAESCBCKEY, "vn_aes_cbc_key", "vn_aes_cbc key info");

struct aes_cbc_keydata {
	struct vncr_generic_params gen;
	u_int32_t seed[4];
	size_t keylen;
	char key[0];
};

struct aes_cbc_key {
	u_int32_t seed[4];
	keyInstance encrypt_key;
	keyInstance decrypt_key;
};

static vncr_create_context_t aes_cbc_create_ctx;
static vncr_delete_context_t aes_cbc_delete_ctx;
static vncr_process_block_t aes_cbc_encrypt;
static vncr_process_block_t aes_cbc_decrypt;

static struct vncr_alg_info aes_cbc_alginfo = {
	VNCRYPT_AES_CBC_ID,
	aes_cbc_create_ctx,
	aes_cbc_delete_ctx,
	aes_cbc_encrypt,
	aes_cbc_decrypt
};

/* looks like CBC encryption mode in rijndael implementation
 * can't use same input and output buffers :(
 */

/* This code depends on 16-byte block length */

static void
cbc_encrypt(keyInstance *key, u_int32_t iv[4], u_int32_t *ptr, size_t len) {
	cipherInstance cipher;
	u_int i;

	if (rijndael_cipherInit(&cipher, MODE_ECB, NULL) != TRUE) {
		printf("vncrypt: rijndael initialization error\n");
		/* plaintext must not escape */
		bzero(ptr, len);
	}

	ptr[0] ^= iv[0];
	ptr[1] ^= iv[1];
	ptr[2] ^= iv[2];
	ptr[3] ^= iv[3];

	for (i = 0; i < len / sizeof(u_int32_t) - 4; i += 4) {
		rijndael_blockEncrypt(&cipher, key, (char *)(&ptr[i]), 16 * 8,
				      (char *)(&ptr[i]));
		ptr[i + 4] ^= ptr[i];
		ptr[i + 5] ^= ptr[i + 1];
		ptr[i + 6] ^= ptr[i + 2];
		ptr[i + 7] ^= ptr[i + 3];
	}
	rijndael_blockEncrypt(&cipher, key, (char *)(&ptr[i]), 16 * 8,
			      (char *)(&ptr[i]));
}

static void
cbc_decrypt(keyInstance *key, u_int32_t iv[4], u_int32_t *ptr, size_t len) {
	cipherInstance cipher;
	u_int i;

	if (rijndael_cipherInit(&cipher, MODE_ECB, NULL) != TRUE) {
		printf("vncrypt: rijndael initialization error\n");
		return ;
	}

	for (i = len / sizeof(u_int32_t) - 4; i >= 4; i -= 4) {
		rijndael_blockDecrypt(&cipher, key, (char *)(&ptr[i]), 16 * 8,
			       	      (char *)(&ptr[i]));
		ptr[i] ^= ptr[i - 4];
		ptr[i + 1] ^= ptr[i - 3];
		ptr[i + 2] ^= ptr[i - 2];
		ptr[i + 3] ^= ptr[i - 1];
	}
	rijndael_blockDecrypt(&cipher, key, (char *)(&ptr[i]), 16 * 8,
		       	      (char *)(&ptr[i]));
	ptr[0] ^= iv[0];
	ptr[1] ^= iv[1];
	ptr[2] ^= iv[2];
	ptr[3] ^= iv[3];
}

static void
checksum(u_int32_t sum[4], const u_int32_t *base, const size_t len) {
	/* simple checksum, simular to ADLER32 */
	u_int i;

        for (i = 0; i < len / sizeof(*base); i++) {
		sum[0] += base[i];
		sum[1] += sum[0];

		sum[2] += base[i];
		sum[3] += sum[2];
        }
}

static void *
aes_cbc_create_ctx(const void *ptr) {
	struct aes_cbc_keydata data;
	struct aes_cbc_key *key;
	u_char *keydata;

	if (copyin(ptr, &data, sizeof(data)))
		return NULL;

	if (data.gen.vncr_algid != VNCRYPT_AES_CBC_ID)
		return NULL;

	MALLOC(keydata, u_char *, data.keylen, M_TEMP, M_WAITOK);
	if (keydata == NULL)
		return NULL;

	if (copyin((const char *)ptr + sizeof(data), keydata, data.keylen)) {
		FREE(keydata, M_TEMP);
		return NULL;
	}

	MALLOC(key, struct aes_cbc_key *, sizeof(struct aes_cbc_key),
	       M_VNAESCBCKEY, M_WAITOK);

	if (key != NULL) {
		key->seed[0] = data.seed[0];
		key->seed[1] = data.seed[1];
		key->seed[2] = data.seed[2];
		key->seed[3] = data.seed[3];
		if (rijndael_makeKey(&key->encrypt_key, DIR_ENCRYPT,
				     data.keylen * 8, keydata) != TRUE ||
		    rijndael_makeKey(&key->decrypt_key, DIR_DECRYPT,
			             data.keylen * 8, keydata) != TRUE) {
			FREE(key, M_VNAESCBCKEY);
			key = NULL;
		}
	}

	FREE(keydata, M_TEMP);

	return key;
}

static void
aes_cbc_delete_ctx(void *ptr) {
	bzero(ptr, sizeof(struct aes_cbc_key));
	FREE(ptr, M_VNAESCBCKEY);
}

/* This code depends on 16-byte block length */

static void
aes_cbc_encrypt(void *cookie, daddr_t offset, void *ptr, size_t size) {
	struct aes_cbc_key *key = (struct aes_cbc_key *)cookie;
	u_int32_t *data = (u_int32_t *)ptr;
	u_int32_t iv[4] = { key->seed[0], key->seed[1],
			    key->seed[2], key->seed[3] };

	iv[0] ^= offset;
	iv[1] ^= offset;
	iv[2] ^= offset;
	iv[3] ^= offset;

	checksum(iv, data + 4, size - sizeof(*data) * 4);

	cbc_encrypt(&key->encrypt_key, iv, ptr, size);
}

static void
aes_cbc_decrypt(void *cookie, daddr_t offset, void *ptr, size_t size) {
	struct aes_cbc_key *key = (struct aes_cbc_key *)cookie;
	u_int32_t *data = (u_int32_t *)ptr;
	u_int32_t iv[4] = { key->seed[0], key->seed[1],
			    key->seed[2], key->seed[3] };

	cbc_decrypt(&key->decrypt_key, data, data + 4,
		    size - sizeof(*data) * 4);

	iv[0] ^= offset;
	iv[1] ^= offset;
	iv[2] ^= offset;
	iv[3] ^= offset;

	checksum(iv, data + 4, size - sizeof(*data) * 4);

	cbc_decrypt(&key->decrypt_key, iv, data, 4 * sizeof(*data));
}

static int
aes_cbc_modevent(module_t mod, int type, void *data) {
	int error = EINVAL;

	switch (type) {
	case MOD_LOAD:
		error = vncrypt_register_alg(&aes_cbc_alginfo);
		break;

	case MOD_UNLOAD:
		/* fall through */
	case MOD_SHUTDOWN:
		error = vncrypt_unregister_alg(VNCRYPT_AES_CBC_ID);
		break;
	}
	return error;
}

static moduledata_t vnaescbcmod = {
        "vnaescbc",
        aes_cbc_modevent,
        0
};

DECLARE_MODULE(vnaescbc, vnaescbcmod, SI_SUB_PSEUDO, SI_ORDER_ANY);
