/*
 * Copyright (c) 2002 Andrey Sverdlichenko
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by 
 *	Andrey Sverdlichenko
 * 4. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Id: vn_aes_multi.c,v 1.4 2002/02/28 14:28:28 rblaze Exp $ */
/* AES (Rijndael) in multikey mode.
 * Each block encrypted on its own subkey derived from single master key.
 * NOTE: the ONLY effect of using different keys for each block is masking
 * dependencies between blocks. It doesn't make your disk as secure as with
 * real independent keys for each block.
 */
#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <dev/vncrypt/vn_crypt.h>
#include <crypto/rijndael/rijndael.h>

#if BITSPERBLOCK != 128
#error Rijndael default block length changed
#endif

MALLOC_DEFINE(M_VNAESMULTIKEY, "vn_aes_multi_key", "vn_aes_multi key info");

struct aes_multi_keydata {
	struct vncr_generic_params gen;
	u_int32_t seed[4];
	size_t subkey_len;
	size_t masterkey_len;
	char masterkey[0];
};

struct aes_multi_key {
	u_int32_t seed[4];
	size_t subkey_len;
	keyInstance key;
};

static vncr_create_context_t aes_multi_create_ctx;
static vncr_delete_context_t aes_multi_delete_ctx;
static vncr_process_block_t aes_multi_encrypt;
static vncr_process_block_t aes_multi_decrypt;

static struct vncr_alg_info aes_multi_alginfo = {
	VNCRYPT_AES_MULTI_ID,
	aes_multi_create_ctx,
	aes_multi_delete_ctx,
	aes_multi_encrypt,
	aes_multi_decrypt
};

/* looks like CBC encryption mode in rijndael implementation
 * can't use same input and output buffers :(
 */

/* This code depends on 16-byte block length */

static void
cbc_encrypt(keyInstance *key, u_int32_t iv[4], u_int32_t *ptr, size_t len) {
	cipherInstance cipher;
	u_int i;

	if (rijndael_cipherInit(&cipher, MODE_ECB, NULL) != TRUE) {
		printf("vncrypt: rijndael initialization error\n");
		/* plaintext must not escape */
		bzero(ptr, len);
	}

	ptr[0] ^= iv[0];
	ptr[1] ^= iv[1];
	ptr[2] ^= iv[2];
	ptr[3] ^= iv[3];

	for (i = 0; i < len / sizeof(u_int32_t) - 4; i += 4) {
		rijndael_blockEncrypt(&cipher, key, (char *)(&ptr[i]), 16 * 8,
				      (char *)(&ptr[i]));
		ptr[i + 4] ^= ptr[i];
		ptr[i + 5] ^= ptr[i + 1];
		ptr[i + 6] ^= ptr[i + 2];
		ptr[i + 7] ^= ptr[i + 3];
	}
	rijndael_blockEncrypt(&cipher, key, (char *)(&ptr[i]), 16 * 8,
			      (char *)(&ptr[i]));
}

static void
cbc_decrypt(keyInstance *key, u_int32_t iv[2], u_int32_t *ptr, size_t len) {
	cipherInstance cipher;
	u_int i;

	if (rijndael_cipherInit(&cipher, MODE_ECB, NULL) != TRUE) {
		printf("vncrypt: rijndael initialization error\n");
		return ;
	}

	for (i = len / sizeof(u_int32_t) - 4; i >= 4; i -= 4) {
		rijndael_blockDecrypt(&cipher, key, (char *)(&ptr[i]), 16 * 8,
			       	      (char *)(&ptr[i]));
		ptr[i] ^= ptr[i - 4];
		ptr[i + 1] ^= ptr[i - 3];
		ptr[i + 2] ^= ptr[i - 2];
		ptr[i + 3] ^= ptr[i - 1];
	}
	rijndael_blockDecrypt(&cipher, key, (char *)(&ptr[i]), 16 * 8,
		       	      (char *)(&ptr[i]));
	ptr[0] ^= iv[0];
	ptr[1] ^= iv[1];
	ptr[2] ^= iv[2];
	ptr[3] ^= iv[3];
}

static void *
aes_multi_create_ctx(const void *ptr) {
	struct aes_multi_keydata data;
	struct aes_multi_key *key;
	u_char *keydata;

	/* copying data from user space */
	if (copyin(ptr, &data, sizeof(data)))
		return NULL;

	if (data.gen.vncr_algid != VNCRYPT_AES_MULTI_ID)
		return NULL;

	MALLOC(keydata, u_char *, data.masterkey_len, M_TEMP, M_WAITOK);
	if (keydata == NULL)
		return NULL;

	if (copyin((const char *)ptr + sizeof(data), keydata,
				data.masterkey_len)) {
		FREE(keydata, M_TEMP);
		return NULL;
	}

	/* generating master key */
	
	MALLOC(key, struct aes_multi_key *, sizeof(struct aes_multi_key),
	       M_VNAESMULTIKEY, M_WAITOK);

	if (key != NULL) {
		key->seed[0] = data.seed[0];
		key->seed[1] = data.seed[1];
		key->seed[2] = data.seed[2];
		key->seed[3] = data.seed[3];
		key->subkey_len = data.subkey_len;
		if (rijndael_makeKey(&key->key, DIR_ENCRYPT,
				     data.masterkey_len * 8, keydata) != TRUE) {
			FREE(key, M_VNAESMULTIKEY);
			key = NULL;
		}
	}

	FREE(keydata, M_TEMP);

	return key;
}

static void
aes_multi_delete_ctx(void *ptr) {
	bzero(ptr, sizeof(struct aes_multi_key));
	FREE(ptr, M_VNAESMULTIKEY);
}

/* This code depends on 16-byte block length */

static int
make_key(keyInstance *master, u_int8_t iv[16], char *keybuf, size_t keylen) {
	cipherInstance cipher;
	int offset = 0;

	bzero(keybuf, keylen);

	if (rijndael_cipherInit(&cipher, MODE_ECB, NULL) != TRUE) {
		printf("vncrypt: rijndael initialization error\n");
		return 1;
	}

	/* I'm afraid that CFB mode can't encrypt to the same buffer too */

	/* FIXME this is SLOW code */
	while (keylen--) {
		if (offset == 0)
			if (rijndael_blockEncrypt(&cipher, master, iv,
					BITSPERBLOCK, iv) != BITSPERBLOCK) {
				printf("vncrypt: rijndael keygen error\n");
				return 1;
			}
		*keybuf ^= iv[offset];
		iv[offset++] = *(keybuf++);

		offset &= 15;
	}
	
	return 0;
}

static void
aes_multi_encrypt(void *cookie, daddr_t offset, void *ptr, size_t size) {
	struct aes_multi_key *key = (struct aes_multi_key *)cookie;
	char subkey[key->subkey_len];
	keyInstance block_key;
	u_int32_t iv[4] = { key->seed[0], key->seed[1],
			    key->seed[2], key->seed[3] };

	/* generating IV */
	iv[0] ^= offset;
	iv[1] ^= offset;
	iv[2] ^= offset;
	iv[3] ^= offset;

	/* making block key */
	if (make_key(&key->key, (char *)iv, subkey, key->subkey_len) ||
	    rijndael_makeKey(&block_key, DIR_ENCRYPT,
				key->subkey_len * 8, subkey) != TRUE) {
		/* plaintext must not escape */
		bzero(ptr, size);
		return;
	}

	/* since block keys are unique, we don't need strong IV */
	bzero(iv, sizeof(iv));

	cbc_encrypt(&block_key, iv, ptr, size);
}

static void
aes_multi_decrypt(void *cookie, daddr_t offset, void *ptr, size_t size) {
	struct aes_multi_key *key = (struct aes_multi_key *)cookie;
	char subkey[key->subkey_len];
	keyInstance block_key;
	u_int32_t iv[4] = { key->seed[0], key->seed[1],
			    key->seed[2], key->seed[3] };

	/* generating IV */
	iv[0] ^= offset;
	iv[1] ^= offset;
	iv[2] ^= offset;
	iv[3] ^= offset;

	/* making block key */
	if (make_key(&key->key, (char *)iv, subkey, key->subkey_len) ||
	    rijndael_makeKey(&block_key, DIR_DECRYPT,
				key->subkey_len * 8, subkey) != TRUE) {
		/* plaintext must not escape */
		bzero(ptr, size);
		return;
	}

	/* since block keys are unique, we don't need strong IV */
	bzero(iv, sizeof(iv));

	cbc_decrypt(&block_key, iv, ptr, size);
}

static int
aes_multi_modevent(module_t mod, int type, void *data) {
	int error = EINVAL;

	switch (type) {
	case MOD_LOAD:
		error = vncrypt_register_alg(&aes_multi_alginfo);
		break;

	case MOD_UNLOAD:
		/* fall through */
	case MOD_SHUTDOWN:
		error = vncrypt_unregister_alg(VNCRYPT_AES_MULTI_ID);
		break;
	}
	return error;
}

static moduledata_t vnaesmultimod = {
        "vnaesmulti",
        aes_multi_modevent,
        0
};

DECLARE_MODULE(vnaesmulti, vnaesmultimod, SI_SUB_PSEUDO, SI_ORDER_ANY);
