/*
 * Copyright (c) 2001 Andrey Sverdlichenko
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by 
 *	Andrey Sverdlichenko
 * 4. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Id: vn_bf_cbc.c,v 1.9 2002/08/21 12:58:58 rblaze Exp $ */
/* Blowfish in CBC mode. IV generated from block number and contents.
 * Expected to be secure.
 * Potential problems:
 * - all disk encrypted on ONE key.
 * - IV uses all but first 8 bytes of block. This will be fixed soon.
 */
#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <dev/vncrypt/vn_crypt.h>
#include <crypto/blowfish/blowfish.h>

#if __FreeBSD_version < 450002
#define vnBF_encrypt(A,B) BF_encrypt(A, B, BF_ENCRYPT)
#define vnBF_decrypt(A,B) BF_encrypt(A, B, BF_DECRYPT)
#else
#define vnBF_encrypt(A,B) BF_encrypt(A, B)
#define vnBF_decrypt(A,B) BF_decrypt(A, B)
#endif


MALLOC_DEFINE(M_VNBFCBCKEY, "vn_bf_cbc_key", "vn_bf_cbc key info");

struct bf_cbc_keydata {
	struct vncr_generic_params gen;
	u_int32_t seed[2];
	size_t keylen;
	char key[0];
};

struct bf_cbc_key {
	u_int32_t seed[2];
	BF_KEY key;
};

static vncr_create_context_t bf_cbc_create_ctx;
static vncr_delete_context_t bf_cbc_delete_ctx;
static vncr_process_block_t bf_cbc_encrypt;
static vncr_process_block_t bf_cbc_decrypt;

static struct vncr_alg_info bf_cbc_alginfo = {
	VNCRYPT_BF_CBC_ID,
	bf_cbc_create_ctx,
	bf_cbc_delete_ctx,
	bf_cbc_encrypt,
	bf_cbc_decrypt
};

static void
cbc_encrypt(BF_KEY *key, u_int32_t iv[2], u_int32_t *ptr, size_t len) {
	u_int i;

	ptr[0] ^= iv[0];
	ptr[1] ^= iv[1];
	
	for (i = 0; i < len / sizeof(u_int32_t) - 2; i += 2) {
		vnBF_encrypt(&ptr[i], key);
		ptr[i + 2] ^= ptr[i];
		ptr[i + 3] ^= ptr[i + 1];
	}
	vnBF_encrypt(&ptr[i], key);
}

static void
cbc_decrypt(BF_KEY *key, u_int32_t iv[2], u_int32_t *ptr, size_t len) {
	u_int i;

	for (i = len / sizeof(u_int32_t) - 2; i >= 2; i -= 2) {
		vnBF_decrypt(&ptr[i], key);
		ptr[i] ^= ptr[i - 2];
		ptr[i + 1] ^= ptr[i - 1];
	}
	vnBF_decrypt(ptr, key);
	ptr[0] ^= iv[0];
	ptr[1] ^= iv[1];
}

static void
checksum(u_int32_t sum[2], const u_int32_t *base, const size_t len) {
	/* simple checksum, simular to ADLER32 */
	u_int i;

        for (i = 0; i < len / sizeof(*base); i++) {
		sum[0] += base[i];
		sum[1] += sum[0];
        }
}

static void *
bf_cbc_create_ctx(const void *ptr) {
	struct bf_cbc_keydata data;
	struct bf_cbc_key *key;
	u_char *keydata;

	if (copyin(ptr, &data, sizeof(data)))
		return NULL;

	if (data.gen.vncr_algid != VNCRYPT_BF_CBC_ID)
		return NULL;

	MALLOC(keydata, u_char *, data.keylen, M_TEMP, M_WAITOK);
	if (keydata == NULL)
		return NULL;

	if (copyin((const char *)ptr + sizeof(data), keydata, data.keylen)) {
		FREE(keydata, M_TEMP);
		return NULL;
	}

	MALLOC(key, struct bf_cbc_key *, sizeof(struct bf_cbc_key),
	       M_VNBFCBCKEY, M_WAITOK);

	if (key != NULL) {
		key->seed[0] = data.seed[0];
		key->seed[1] = data.seed[1];
		BF_set_key(&key->key, data.keylen, keydata);
	}

	FREE(keydata, M_TEMP);

	return key;
}

static void
bf_cbc_delete_ctx(void *ptr) {
	bzero(ptr, sizeof(struct bf_cbc_key));
	FREE(ptr, M_VNBFCBCKEY);
}

static void
bf_cbc_encrypt(void *cookie, daddr_t offset, void *ptr, size_t size) {
	struct bf_cbc_key *key = (struct bf_cbc_key *)cookie;
	u_int32_t *data = (u_int32_t *)ptr;
	u_int32_t iv[2] = { key->seed[0], key->seed[1] };

	iv[0] ^= offset;
	iv[1] ^= offset;

	checksum(iv, data + 2, size - sizeof(*data) * 2);

	cbc_encrypt(&key->key, iv, ptr, size);
}

static void
bf_cbc_decrypt(void *cookie, daddr_t offset, void *ptr, size_t size) {
	struct bf_cbc_key *key = (struct bf_cbc_key *)cookie;
	u_int32_t *data = (u_int32_t *)ptr;
	u_int32_t iv[2] = { key->seed[0], key->seed[1] };

	iv[0] ^= offset;
	iv[1] ^= offset;

	cbc_decrypt(&key->key, data, data + 2, size - sizeof(*data) * 2);

	checksum(iv, data + 2, size - sizeof(*data) * 2);

	cbc_decrypt(&key->key, iv, data, 2 * sizeof(*data));
}

static int
bf_cbc_modevent(module_t mod, int type, void *data) {
	int error = EINVAL;

	switch (type) {
	case MOD_LOAD:
		error = vncrypt_register_alg(&bf_cbc_alginfo);
		break;

	case MOD_UNLOAD:
		/* fall through */
	case MOD_SHUTDOWN:
		error = vncrypt_unregister_alg(VNCRYPT_BF_CBC_ID);
		break;
	}
	return error;
}

static moduledata_t vnbfcbcmod = {
        "vnbfcbc",
        bf_cbc_modevent,
        0
};

DECLARE_MODULE(vnbfcbc, vnbfcbcmod, SI_SUB_PSEUDO, SI_ORDER_ANY);
