/*
 * Copyright (c) 2001 Andrey Sverdlichenko
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by 
 *	Andrey Sverdlichenko
 * 4. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Id: vn_crypt.c,v 1.5 2001/12/07 10:10:57 rblaze Exp $ */
#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <dev/vncrypt/vn_crypt.h>

MALLOC_DEFINE(M_VNCRYPTKEY, "vncrypt_key", "vncrypt key info");

/*    */
struct ctx_info {
	struct vncr_alg_info *alg;
	void *params;
	int active;
	int timeout_action;
	long timeout;
	long last_use;
};

static LIST_HEAD(, vncr_alg_info) algs = LIST_HEAD_INITIALIZER(bogus);

int
vncrypt_register_alg(struct vncr_alg_info *alginfo) {
	struct vncr_alg_info *ptr;

	LIST_FOREACH(ptr, &algs, chain)
		if (ptr->alg_id == alginfo->alg_id)
			return EEXIST;

	LIST_INSERT_HEAD(&algs, alginfo, chain);

	return 0;
}

int
vncrypt_unregister_alg(int id) {
	struct vncr_alg_info *ptr;

	LIST_FOREACH(ptr, &algs, chain)
		if (ptr->alg_id == id) {
			LIST_REMOVE(ptr, chain);
			return 0;
		}

	return ESRCH;
}

int
vncrypt_enable(const void *ptr, void **cookie) {
	const char *dptr = *(const char **)ptr;
	struct ctx_info *ctx;
	struct vncr_alg_info *alg;
	int error;
	struct vncr_generic_params params;
	struct timeval tv;

	error = copyin(dptr, &params, sizeof(params));
	if (error)
		return error;

	LIST_FOREACH(alg, &algs, chain)
		if (alg->alg_id == params.vncr_algid)
			break;

	if (alg == NULL)
		return ESRCH;

	MALLOC(ctx, struct ctx_info *, sizeof(struct ctx_info), M_VNCRYPTKEY,
	       M_WAITOK);

	if (ctx == NULL)
		return ENOMEM;

	ctx->params = alg->alg_create_ctx(dptr);

	if (ctx->params == NULL) {
		FREE(ctx, M_VNCRYPTKEY);
		return EINVAL;
	}


	getmicrouptime(&tv);

	ctx->alg = alg;
	ctx->active = 1;
	ctx->timeout_action = params.vncr_timeout_action;
	ctx->timeout = params.vncr_timeout;
	ctx->last_use = tv.tv_sec;
	
	*cookie = ctx;

	return 0;
}

void
vncrypt_delete(void *cookie) {
	struct ctx_info *ctx = (struct ctx_info *)cookie;

	ctx->alg->alg_delete_ctx(ctx->params);
	FREE(ctx, M_VNCRYPTKEY);
}

void
vncrypt_encrypt_block(void *cookie, daddr_t offset, void *ptr, size_t size) {
	struct ctx_info *ctx = (struct ctx_info *)cookie;

	if (ctx->active) {
		ctx->alg->alg_encrypt(ctx->params, offset, ptr, size);
	} else {
		/* Plaintext shouldn't go anywhere,
		 * we better lose this data
		 */
		bzero(ptr, size);
	}
}

void
vncrypt_decrypt_block(void *cookie, daddr_t offset, void *ptr, size_t size) {
	struct ctx_info *ctx = (struct ctx_info *)cookie;

	if (ctx->active)
		ctx->alg->alg_decrypt(ctx->params, offset, ptr, size);
}

int
vncrypt_check(void *cookie) {
	struct ctx_info *ctx = (struct ctx_info *)cookie;
	struct timeval tv;

	if (ctx->timeout == 0)
		return 0;

	if (!ctx->active)
		return ctx->timeout_action;

	getmicrouptime(&tv);

	if (tv.tv_sec > ctx->last_use + ctx->timeout) {
		ctx->active = 0;
		return ctx->timeout_action;
	}

	ctx->last_use = tv.tv_sec;

	return 0;
}
